import pytest

from cg.constants.constants import MicrosaltAppTags
from cg.meta.workflow.microsalt.metrics_parser.models import (
    MicrosaltSamtoolsStats,
    PicardMarkduplicate,
    SampleMetrics,
)
from cg.meta.workflow.microsalt.quality_controller.models import (
    CaseQualityResult,
    SampleQualityResult,
)
from cg.meta.workflow.microsalt.quality_controller.quality_controller import (
    MicroSALTQualityController,
)
from cg.store.store import Store


def create_sample_metrics(
    total_reads: int | None = 100,
    mapped_rate: float | None = 0.8,
    duplication_rate: float | None = 0.1,
    insert_size: int | None = 200,
    average_coverage: float | None = 30.0,
    coverage_10x: float | None = 95.0,
) -> SampleMetrics:
    return SampleMetrics(
        microsalt_samtools_stats=MicrosaltSamtoolsStats(
            total_reads=total_reads,
            mapped_rate=mapped_rate,
            average_coverage=average_coverage,
            coverage_10x=coverage_10x,
        ),
        picard_markduplicate=PicardMarkduplicate(
            insert_size=insert_size, duplication_rate=duplication_rate
        ),
    )


def create_quality_result(
    sample_id: str = "sample_1",
    passes_qc: bool = True,
    is_control: bool = False,
    application_tag: str = MicrosaltAppTags.MWRNXTR003,
    passes_reads_qc: bool = True,
    passes_mapping_qc: bool = True,
    passes_duplication_qc: bool = True,
    passes_inserts_qc: bool = True,
    passes_coverage_qc: bool = True,
    passes_10x_coverage_qc: bool = True,
) -> SampleQualityResult:
    return SampleQualityResult(
        sample_id=sample_id,
        passes_qc=passes_qc,
        is_control=is_control,
        application_tag=application_tag,
        passes_reads_qc=passes_reads_qc,
        passes_mapping_qc=passes_mapping_qc,
        passes_duplication_qc=passes_duplication_qc,
        passes_inserts_qc=passes_inserts_qc,
        passes_coverage_qc=passes_coverage_qc,
        passes_10x_coverage_qc=passes_10x_coverage_qc,
    )


@pytest.fixture
def quality_results() -> list[SampleQualityResult]:
    return [
        SampleQualityResult(
            sample_id="sample_1",
            passes_qc=False,
            is_control=True,
            application_tag=MicrosaltAppTags.MWRNXTR003,
            passes_reads_qc=True,
            passes_mapping_qc=True,
            passes_duplication_qc=False,
            passes_inserts_qc=True,
            passes_coverage_qc=True,
            passes_10x_coverage_qc=True,
        ),
        SampleQualityResult(
            sample_id="sample_2",
            passes_qc=True,
            is_control=False,
            application_tag=MicrosaltAppTags.MWRNXTR003,
            passes_reads_qc=True,
            passes_mapping_qc=True,
            passes_duplication_qc=True,
            passes_inserts_qc=True,
            passes_coverage_qc=True,
            passes_10x_coverage_qc=True,
        ),
        SampleQualityResult(
            sample_id="sample_3",
            passes_qc=False,
            is_control=False,
            application_tag=MicrosaltAppTags.MWRNXTR003,
            passes_reads_qc=False,
            passes_mapping_qc=True,
            passes_duplication_qc=False,
            passes_inserts_qc=True,
            passes_coverage_qc=True,
            passes_10x_coverage_qc=False,
        ),
    ]


@pytest.fixture
def case_result():
    return CaseQualityResult(
        passes_qc=False,
        control_passes_qc=True,
        urgent_passes_qc=True,
        non_urgent_passes_qc=True,
    )


@pytest.fixture
def quality_controller(store: Store) -> MicroSALTQualityController:
    return MicroSALTQualityController(store)



WORKFLOW_TEMPLATE = """#! /bin/bash
#SBATCH --job-name={job_name}
#SBATCH --account={account}
#SBATCH --ntasks={number_tasks}
#SBATCH --mem={memory}G
#SBATCH --error={output_fastq_dir}/{job_name}.stderr
#SBATCH --output={output_fastq_dir}/{job_name}.stdout
#SBATCH --mail-type=ALL
#SBATCH --mail-user={email}
#SBATCH --time={time}
#SBATCH --qos={quality_of_service}

set -eu -o pipefail
log() {{
    NOW=$(date +"%Y-%m-%dT%H:%M:%S")
    echo "[${{NOW}}] $*" 1>&2;
}}
log "Running on: $(hostname)"

log "Executing: {bin_downsample} {input_fastq_dir}/ {output_fastq_dir}/ {downsample_to} {original_sample_reads}"
{bin_downsample} {input_fastq_dir}/ {output_fastq_dir}/ {downsample_to} {original_sample_reads}
"""



"""Module for modeling flow cells."""

import datetime
import logging
import os
from pathlib import Path
from typing import Type

from typing_extensions import Literal

from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import get_latest_analysis_path
from cg.constants.constants import LENGTH_LONG_DATE
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.sequencing import SEQUENCER_TYPES, Sequencers
from cg.constants.symbols import EMPTY_STRING
from cg.exc import FlowCellError
from cg.models.demultiplex.run_parameters import (
    RunParameters,
    RunParametersHiSeq,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)
from cg.models.run_devices.utils import parse_date
from cg.utils.files import get_source_creation_time_stamp
from cg.utils.time import format_time_from_ctime

LOG = logging.getLogger(__name__)
RUN_PARAMETERS_CONSTRUCTOR: dict[str, Type] = {
    Sequencers.HISEQGA: RunParametersHiSeq,
    Sequencers.HISEQX: RunParametersHiSeq,
    Sequencers.NOVASEQ: RunParametersNovaSeq6000,
    Sequencers.NOVASEQX: RunParametersNovaSeqX,
}


class IlluminaRunDirectoryData:
    """Class to collect information about sequencing run directories and their particular files."""

    def __init__(self, sequencing_run_path: Path):
        LOG.debug(f"Instantiating IlluminaRunDirectoryData with path {sequencing_run_path}")
        self.path: Path = sequencing_run_path
        self.machine_name: str = EMPTY_STRING
        self._run_parameters: RunParameters | None = None
        self.run_date: datetime.datetime = datetime.datetime.now()
        self.machine_number: int = 0
        self.base_name: str = EMPTY_STRING  # Base name is flow cell-id + flow cell position
        self.id: str = EMPTY_STRING
        self.position: Literal["A", "B"] = "A"
        self.parse_sequencing_run_dir_name()
        self._sample_sheet_path_hk: Path | None = None
        self.sample_sheet_validator = SampleSheetValidator()

    def parse_sequencing_run_dir_name(self):
        """Parse relevant information from sequencing run name.
        This will assume that the sequencing run naming convention is used. If not we skip the sequencing run.
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>
        Example: '230912_A00187_1009_AHK33MDRX3'.
        """

        self.validate_sequencing_run_dir_name()
        self.run_date = self._parse_date()
        self.machine_name = self.split_sequencing_run_name[1]
        self.machine_number = int(self.split_sequencing_run_name[2])
        base_name: str = self.split_sequencing_run_name[-1]
        self.base_name = base_name
        LOG.debug(f"Set sequencing run id to {base_name}")
        self.id = base_name[1:]
        self.position = base_name[0]

    @property
    def split_sequencing_run_name(self) -> list[str]:
        """Return split sequencing run name."""
        return self.path.name.split("_")

    @property
    def full_name(self) -> str:
        """Return sequencing run full name."""
        return self.path.name

    @property
    def sequenced_at(self) -> datetime:
        """Return the sequencing date for the sequencing run."""
        date_part: str = self.full_name.split("_")[0]
        return parse_date(date_part)

    @property
    def sample_sheet_path(self) -> Path:
        """
        Return sample sheet path.
        """
        return Path(
            self.get_sequencing_runs_dir(), DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
        )

    @property
    def get_sequencing_completed_path(self) -> Path | None:
        """Return the path to the sequencing completed file."""
        sequencing_completed_paths: dict = {
            Sequencers.NOVASEQ: Path(
                self.get_sequencing_runs_dir(),
                DemultiplexingDirsAndFiles.SEQUENCE_COMPLETED,
            ),
            Sequencers.HISEQX: Path(
                self.get_sequencing_runs_dir(),
                DemultiplexingDirsAndFiles.SEQUENCING_COMPLETED,
            ),
        }
        return (
            sequencing_completed_paths[self.sequencer_type]
            if self.sequencer_type in sequencing_completed_paths.keys()
            else None
        )

    def set_sample_sheet_path_hk(self, hk_path: Path):
        self._sample_sheet_path_hk = hk_path

    def get_sample_sheet_path_hk(self) -> Path:
        if not self._sample_sheet_path_hk:
            raise FlowCellError("Attribute _sample_sheet_path_hk has not been assigned yet")
        return self._sample_sheet_path_hk

    def get_sequencing_runs_dir(self) -> Path:
        """
        Return the flow cells run directory regardless of the path used to initialise the IlluminaRunDirectoryData.
        """
        current_path: str = self.path.as_posix()
        if DemultiplexingDirsAndFiles.DEMULTIPLEXED_RUNS_DIRECTORY_NAME in current_path:
            return Path(
                str.replace(
                    current_path,
                    DemultiplexingDirsAndFiles.DEMULTIPLEXED_RUNS_DIRECTORY_NAME,
                    DemultiplexingDirsAndFiles.SEQUENCING_RUNS_DIRECTORY_NAME,
                )
            )
        return self.path

    def get_demultiplexed_runs_dir(self) -> Path:
        """
        Return the demultiplexed run directory regardless of the path used to initialise the IlluminaRunDirectoryData.
        """
        current_path: str = self.path.as_posix()
        if DemultiplexingDirsAndFiles.SEQUENCING_RUNS_DIRECTORY_NAME in current_path:
            return Path(
                str.replace(
                    current_path,
                    DemultiplexingDirsAndFiles.SEQUENCING_RUNS_DIRECTORY_NAME,
                    DemultiplexingDirsAndFiles.DEMULTIPLEXED_RUNS_DIRECTORY_NAME,
                )
            )
        return self.path

    @property
    def run_parameters_path(self) -> Path:
        """Return path to run parameters file if it exists.
        Raises:
            FlowCellError if the sequencing run has no run parameters file."""
        flow_cell_run_dir: Path = self.get_sequencing_runs_dir()
        if DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE in os.listdir(flow_cell_run_dir):
            return Path(flow_cell_run_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE)
        elif DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE in os.listdir(flow_cell_run_dir):
            return Path(flow_cell_run_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE)
        else:
            message: str = f"No run parameters file found in sequencing run {flow_cell_run_dir}"
            LOG.error(message)
            raise FlowCellError(message)

    @property
    def run_parameters(self) -> RunParameters:
        """Return run parameters object."""
        if not self._run_parameters:
            self._run_parameters = RUN_PARAMETERS_CONSTRUCTOR[self.sequencer_type](
                run_parameters_path=self.run_parameters_path
            )
        return self._run_parameters

    @property
    def sequencer_type(
        self,
    ) -> Literal[Sequencers.HISEQX, Sequencers.HISEQGA, Sequencers.NOVASEQ, Sequencers.NOVASEQX]:
        """Return the sequencer type."""
        return SEQUENCER_TYPES[self.machine_name]

    @property
    def rta_complete_path(self) -> Path:
        """Return RTAComplete path."""
        return Path(self.get_sequencing_runs_dir(), DemultiplexingDirsAndFiles.RTACOMPLETE)

    @property
    def copy_complete_path(self) -> Path:
        """Return copy complete path."""
        return Path(self.get_sequencing_runs_dir(), DemultiplexingDirsAndFiles.COPY_COMPLETE)

    @property
    def demultiplexing_started_path(self) -> Path:
        """Return demux started path."""
        flow_cell_run_dir: Path = self.get_sequencing_runs_dir()
        return Path(flow_cell_run_dir, DemultiplexingDirsAndFiles.DEMUX_STARTED)

    @property
    def demultiplex_software_info_path(self) -> Path:
        """Return demultiplex software info path."""
        demux_run_dir = self.get_demultiplexed_runs_dir()
        return Path(demux_run_dir, DemultiplexingDirsAndFiles.DEMUX_VERSION_FILE)

    @property
    def demux_complete_path(self) -> Path:
        """Return demux complete path."""
        demux_run_dir = self.get_demultiplexed_runs_dir()
        return Path(demux_run_dir, DemultiplexingDirsAndFiles.DEMUX_COMPLETE)

    @property
    def trailblazer_config_path(self) -> Path:
        """Return file to SLURM job ids path."""
        return Path(self.get_sequencing_runs_dir(), "slurm_job_ids.yaml")

    @property
    def is_demultiplexing_complete(self) -> bool:
        return Path(self.demux_complete_path).exists()

    def _parse_date(self):
        """Return the parsed date in the correct format."""
        if len(self.split_sequencing_run_name[0]) == LENGTH_LONG_DATE:
            return datetime.datetime.strptime(self.split_sequencing_run_name[0], "%Y%m%d")
        return datetime.datetime.strptime(self.split_sequencing_run_name[0], "%y%m%d")

    def validate_sequencing_run_dir_name(self) -> None:
        """
        Validate on the following criteria:
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>
        Example: '230912_A00187_1009_AHK33MDRX3'.
        """
        if len(self.split_sequencing_run_name) != 4:
            message = (
                f"Flowcell {self.full_name} does not follow the sequencing run naming convention"
            )
            LOG.warning(message)
            raise FlowCellError(message)

    def has_demultiplexing_started_locally(self) -> bool:
        """Check if demultiplexing has started path exists on the cluster."""
        return self.demultiplexing_started_path.exists()

    def has_demultiplexing_started_on_sequencer(self) -> bool:
        """Check if demultiplexing has started on the NovaSeqX machine."""
        latest_analysis: Path = get_latest_analysis_path(self.get_demultiplexed_runs_dir())
        if not latest_analysis:
            return False
        return Path(
            latest_analysis, DemultiplexingDirsAndFiles.DATA, DemultiplexingDirsAndFiles.BCL_CONVERT
        ).exists()

    def sample_sheet_exists(self) -> bool:
        """Check if sample sheet exists."""
        LOG.info("Check if sample sheet exists")
        return self.sample_sheet_path.exists()

    @property
    def sample_sheet(self) -> SampleSheet:
        """Return sample sheet object."""
        if not self._sample_sheet_path_hk:
            raise FlowCellError("Sample sheet path has not been assigned yet")
        return self.sample_sheet_validator.get_sample_sheet_object_from_file(
            file_path=self._sample_sheet_path_hk
        )

    def is_sequencing_done(self) -> bool:
        """Check if sequencing is done.
        This is indicated by that the file RTAComplete.txt exists.
        """
        LOG.info("Check if sequencing is done")
        return self.rta_complete_path.exists()

    def is_copy_completed(self) -> bool:
        """Check if copy of sequencing run is done.
        This is indicated by that the file CopyComplete.txt exists.
        """
        LOG.info("Check if copy of data from sequence instrument is ready")
        return self.copy_complete_path.exists()

    def is_sequencing_run_ready(self) -> bool:
        """Check if a sequencing run is ready for downstream processing.

        A sequencing run is ready if the two files RTAComplete.txt and CopyComplete.txt exist in the
        sequencing run directory.
        """
        LOG.info("Check if sequencing run is ready for downstream processing")
        if not self.is_sequencing_done():
            LOG.warning(f"Sequencing is not completed for sequencing run {self.id}")
            return False
        LOG.debug(f"Sequence is done for sequencing run {self.id}")
        if not self.is_copy_completed():
            LOG.warning(f"Copy of sequence data is not ready for sequencing run {self.id}")
            return False
        LOG.debug(f"All data has been transferred for sequencing run {self.id}")
        LOG.info(f"Sequencing run {self.id} is ready for downstream processing")
        return True

    def get_run_completion_status(self) -> Path | None:
        """Return the run completion status path."""
        flow_cells_dir: Path = self.get_sequencing_runs_dir()
        file_path = Path(flow_cells_dir, DemultiplexingDirsAndFiles.RUN_COMPLETION_STATUS)
        if file_path.exists():
            return file_path
        return None

    @property
    def demultiplexing_started_at(self) -> datetime.datetime | None:
        """Get the demultiplexing started time stamp from the sequencing run dir."""
        try:
            time: float = get_source_creation_time_stamp(self.demultiplexing_started_path)
            return format_time_from_ctime(time)
        except FileNotFoundError as error:
            LOG.warning(error)
            return None

    @property
    def demultiplexing_completed_at(self) -> datetime.datetime | None:
        """Get the demultiplexing completed time stamp from the demultiplexed runs dir."""
        try:
            time: float = get_source_creation_time_stamp(self.demux_complete_path)
            return format_time_from_ctime(time)
        except FileNotFoundError as error:
            LOG.warning(error)
            return None

    def __str__(self):
        return f"FlowCell(path={self.path},run_parameters_path={self.run_parameters_path})"


def get_sequencing_runs_from_path(sequencing_run_dir: Path) -> list[IlluminaRunDirectoryData]:
    """Return sequencing run objects from sequencing run dir."""
    sequencing_runs: list[IlluminaRunDirectoryData] = []
    LOG.debug(f"Search for flow cells ready to encrypt in {sequencing_run_dir}")
    for run_dir in sequencing_run_dir.iterdir():
        if not run_dir.is_dir():
            continue
        LOG.debug(f"Found directory: {run_dir}")
        try:
            sequencing_run = IlluminaRunDirectoryData(sequencing_run_path=run_dir)
        except FlowCellError:
            continue
        sequencing_runs.append(sequencing_run)
    return sequencing_runs



from cg.server.dto.sequencing_metrics.sequencing_metrics_request import SequencingMetricsRequest
from cg.store.models import IlluminaSampleSequencingMetrics


def parse_metrics_into_request(
    metrics: list[IlluminaSampleSequencingMetrics],
) -> list[SequencingMetricsRequest]:
    """Parse sequencing metrics into a request."""
    parsed_metrics: list[SequencingMetricsRequest] = []
    for metric in metrics:
        parsed_metric = SequencingMetricsRequest(
            flow_cell_name=metric.instrument_run.device.internal_id,
            flow_cell_lane_number=metric.flow_cell_lane,
            sample_internal_id=metric.sample.internal_id,
            sample_total_reads_in_lane=metric.total_reads_in_lane,
            sample_base_percentage_passing_q30=metric.base_passing_q30_percent,
        )
        parsed_metrics.append(parsed_metric)
    return parsed_metrics



from decimal import Decimal
from json import JSONEncoder

from flask_admin import Admin
from flask_cors import CORS
from flask_wtf.csrf import CSRFProtect

from cg.apps.lims import LimsAPI
from cg.apps.osticket import OsTicket
from cg.apps.tb.api import TrailblazerAPI
from cg.services.delivery_message.delivery_message_service import DeliveryMessageService
from cg.services.orders.order_service.order_service import OrderService
from cg.services.orders.order_summary_service.order_summary_service import (
    OrderSummaryService,
)
from cg.services.sample_service.sample_service import SampleService
from cg.store.database import initialize_database
from cg.store.store import Store


class FlaskLims(LimsAPI):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        config = {
            "lims": {
                "host": app.config["lims_host"],
                "username": app.config["lims_username"],
                "password": app.config["lims_password"],
            }
        }
        super(FlaskLims, self).__init__(config)


class FlaskStore(Store):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        uri = app.config["cg_sql_database_uri"]
        initialize_database(uri)
        super(FlaskStore, self).__init__()


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)
        return super().default(obj)


class AnalysisClient(TrailblazerAPI):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        service_account: str = app.config["trailblazer_service_account"]
        service_account_auth_file: str = app.config["trailblazer_service_account_auth_file"]
        host: str = app.config["trailblazer_host"]
        config = {
            "trailblazer": {
                "service_account": service_account,
                "service_account_auth_file": service_account_auth_file,
                "host": host,
            }
        }
        super(AnalysisClient, self).__init__(config)


cors = CORS(resources={r"/api/*": {"origins": "*"}}, supports_credentials=True)
csrf = CSRFProtect()
db = FlaskStore()

admin = Admin(name="Clinical Genomics")
lims = FlaskLims()
osticket = OsTicket()
analysis_client = AnalysisClient()
delivery_message_service = DeliveryMessageService(store=db, trailblazer_api=analysis_client)
summary_service = OrderSummaryService(store=db, analysis_client=analysis_client)
order_service = OrderService(store=db, status_service=summary_service)
sample_service = SampleService(db)



"""Constants related to demultiplexing."""

from enum import Enum, StrEnum

from pydantic import BaseModel

from cg.constants.sequencing import Sequencers


class DemultiplexingDirsAndFiles(StrEnum):
    """Demultiplexing related directories and files."""

    COPY_COMPLETE: str = "CopyComplete.txt"
    DELIVERY: str = "delivery.txt"
    DEMUX_STARTED: str = "demuxstarted.txt"
    DEMUX_COMPLETE: str = "demuxcomplete.txt"
    RTACOMPLETE: str = "RTAComplete.txt"
    RUN_PARAMETERS_PASCAL_CASE: str = "RunParameters.xml"
    RUN_PARAMETERS_CAMEL_CASE: str = "runParameters.xml"
    SAMPLE_SHEET_FILE_NAME: str = "SampleSheet.csv"
    UNALIGNED_DIR_NAME: str = "Unaligned"
    QUEUED_FOR_POST_PROCESSING: str = "post_processing_queued.txt"
    ANALYSIS_COMPLETED: str = "Secondary_Analysis_Complete.txt"
    ANALYSIS: str = "Analysis"
    DATA: str = "Data"
    BCL_CONVERT: str = "BCLConvert"
    SEQUENCING_RUNS_DIRECTORY_NAME: str = "sequencing-runs"
    DEMULTIPLEXED_RUNS_DIRECTORY_NAME: str = "demultiplexed-runs"
    ILLUMINA_FILE_MANIFEST: str = "Manifest.tsv"
    CG_FILE_MANIFEST: str = "file_manifest.tsv"
    INTER_OP: str = "InterOp"
    RUN_COMPLETION_STATUS: str = "RunCompletionStatus.xml"
    DEMUX_VERSION_FILE: str = "dragen-replay.json"
    SEQUENCING_COMPLETED: str = "SequencingComplete.txt"
    SEQUENCE_COMPLETED: str = "SequenceComplete.txt"


class RunParametersXMLNodes(StrEnum):
    """Define names of the used XML nodes in run parameters files."""

    # Node names
    APPLICATION: str = ".Application"
    APPLICATION_NAME: str = ".//ApplicationName"
    APPLICATION_VERSION: str = ".ApplicationVersion"
    CYCLES: str = "Cycles"
    INDEX_1_HISEQ: str = ".//IndexRead1"
    INDEX_2_HISEQ: str = ".//IndexRead2"
    INDEX_1_NOVASEQ_6000: str = "./IndexRead1NumberOfCycles"
    INDEX_2_NOVASEQ_6000: str = "./IndexRead2NumberOfCycles"
    INDEX_1_NOVASEQ_X: str = "Index1"
    INDEX_2_NOVASEQ_X: str = "Index2"
    INNER_READ: str = ".//Read"
    INSTRUMENT_TYPE: str = ".InstrumentType"
    PLANNED_READS_HISEQ: str = ".//Reads"
    PLANNED_READS_NOVASEQ_X: str = "./PlannedReads"
    READ_1_HISEQ: str = ".//Read1"
    READ_2_HISEQ: str = ".//Read2"
    READ_1_NOVASEQ_6000: str = "./Read1NumberOfCycles"
    READ_2_NOVASEQ_6000: str = "./Read2NumberOfCycles"
    READ_1_NOVASEQ_X: str = "Read1"
    READ_2_NOVASEQ_X: str = "Read2"
    READ_NAME: str = "ReadName"
    REAGENT_KIT_VERSION: str = "./RfidsInfo/SbsConsumableVersion"
    SEQUENCER_ID: str = ".//ScannerID"
    FLOW_CELL_MODE: str = ".//FlowCellMode"
    MODE: str = ".//Mode"

    # Node Values
    HISEQ_APPLICATION: str = "HiSeq Control Software"
    NOVASEQ_6000_APPLICATION: str = "NovaSeq Control Software"
    NOVASEQ_X_INSTRUMENT: str = "NovaSeqXPlus"
    UNKNOWN_REAGENT_KIT_VERSION: str = "unknown"


class SampleSheetBcl2FastqSections:
    """Class with all necessary constants for building a NovaSeqX sample sheet."""

    class Data(StrEnum):
        HEADER: str = "[Data]"
        FLOW_CELL_ID: str = "FCID"
        LANE: str = "Lane"
        SAMPLE_INTERNAL_ID_BCL2FASTQ: str = "SampleID"
        SAMPLE_REFERENCE: str = "SampleRef"
        INDEX_1: str = "index"
        INDEX_2: str = "index2"
        SAMPLE_NAME: str = "SampleName"
        CONTROL: str = "Control"
        RECIPE: str = "Recipe"
        OPERATOR: str = "Operator"
        SAMPLE_PROJECT_BCL2FASTQ: str = "Project"


class SampleSheetBCLConvertSections:
    """Class with all necessary constants for building a version 2 sample sheet."""

    class Header(StrEnum):
        HEADER: str = "[Header]"
        RUN_NAME: str = "RunName"
        INSTRUMENT_PLATFORM_TITLE: str = "InstrumentPlatform"
        INDEX_SETTINGS: str = "IndexSettings"

        @classmethod
        def file_format(cls) -> list[str]:
            return ["FileFormatVersion", "2"]

        @classmethod
        def instrument_platform_sequencer(cls) -> dict[str, str]:
            return {
                Sequencers.NOVASEQ: "NovaSeq6000",
                Sequencers.NOVASEQX: "NovaSeqXSeries",
                Sequencers.HISEQX: "HiSeqXSeries",
                Sequencers.HISEQGA: "HiSeq2500",
            }

        @classmethod
        def index_orientation_forward(cls) -> list[str]:
            return ["IndexOrientation", "Forward"]

    class Reads(StrEnum):
        HEADER: str = "[Reads]"
        READ_CYCLES_1: str = "Read1Cycles"
        READ_CYCLES_2: str = "Read2Cycles"
        INDEX_CYCLES_1: str = "Index1Cycles"
        INDEX_CYCLES_2: str = "Index2Cycles"

    class Settings(StrEnum):
        HEADER: str = "[BCLConvert_Settings]"

        @classmethod
        def software_version(cls) -> list[str]:
            return ["SoftwareVersion", "4.1.7"]

        @classmethod
        def fastq_compression_format(cls) -> list[str]:
            return ["FastqCompressionFormat", "gzip"]

    class Data(StrEnum):
        HEADER: str = "[BCLConvert_Data]"
        LANE: str = "Lane"
        SAMPLE_INTERNAL_ID: str = "Sample_ID"
        INDEX_1: str = "Index"
        INDEX_2: str = "Index2"
        OVERRIDE_CYCLES: str = "OverrideCycles"
        BARCODE_MISMATCHES_1: str = "BarcodeMismatchesIndex1"
        BARCODE_MISMATCHES_2: str = "BarcodeMismatchesIndex2"

        @classmethod
        def column_names(cls) -> list[str]:
            return [
                cls.LANE,
                cls.SAMPLE_INTERNAL_ID,
                cls.INDEX_1,
                cls.INDEX_2,
                cls.OVERRIDE_CYCLES,
                cls.BARCODE_MISMATCHES_1,
                cls.BARCODE_MISMATCHES_2,
            ]


class IndexOverrideCycles(StrEnum):
    """Class with the possible values that index cycles can take."""

    FULL_10_INDEX: str = "I10;"
    FULL_8_INDEX: str = "I8;"
    IGNORED_10_INDEX: str = "N10;"
    IGNORED_8_INDEX: str = "N8;"
    INDEX_8_IGNORED_2: str = "I8N2;"
    INDEX_8_IGNORED_2_REVERSED: str = "N2I8;"


CUSTOM_INDEX_TAIL = "NNNNNNNNN"
FASTQ_FILE_SUFFIXES: list[str] = [".fastq", ".gz"]
UNDETERMINED: str = "Undetermined"

NEW_NOVASEQ_CONTROL_SOFTWARE_VERSION: str = "1.7.0"
NEW_NOVASEQ_REAGENT_KIT_VERSION: str = "1.5"

FORWARD_INDEX_CYCLE_PATTERN: str = r"I(\d+)N(\d+)"
REVERSE_INDEX_CYCLE_PATTERN: str = r"N(\d+)I(\d+)"


class IndexSettings(BaseModel):
    """
    Holds the settings defining how the sample indexes should be handled in the sample sheet.
    These vary between machines and versions.

        Attributes:
            should_i5_be_reverse_complemented (bool): Whether the i5 index should be reverse complemented.
            are_i5_override_cycles_reverse_complemented (bool): Whether the override cycles for i5 should be written in as NXIX.

    """

    name: str
    should_i5_be_reverse_complemented: bool
    are_i5_override_cycles_reverse_complemented: bool


# The logic for the settings below are acquired empirically, any changes should be well motivated
# and rigorously tested.

NOVASEQ_X_INDEX_SETTINGS = IndexSettings(
    name="NovaSeqX",
    should_i5_be_reverse_complemented=False,
    are_i5_override_cycles_reverse_complemented=True,
)
NOVASEQ_6000_POST_1_5_KITS_INDEX_SETTINGS = IndexSettings(
    name="NovaSeq6000Post1.5Kits",
    should_i5_be_reverse_complemented=True,
    are_i5_override_cycles_reverse_complemented=False,
)
NO_REVERSE_COMPLEMENTS_INDEX_SETTINGS = IndexSettings(
    name="NoReverseComplements",
    should_i5_be_reverse_complemented=False,
    are_i5_override_cycles_reverse_complemented=False,
)

NAME_TO_INDEX_SETTINGS: dict[str, IndexSettings] = {
    "NovaSeqX": NOVASEQ_X_INDEX_SETTINGS,
    "NovaSeq6000Post1.5Kits": NOVASEQ_6000_POST_1_5_KITS_INDEX_SETTINGS,
    "NoReverseComplements": NO_REVERSE_COMPLEMENTS_INDEX_SETTINGS,
}


class RunCompletionStatusNodes(StrEnum):
    RUN_START: str = ".//RunStartTime"
    RUN_END: str = ".//RunEndTime"



"""Test observations API methods."""

import logging
from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from pytest_mock import MockFixture

from cg.apps.lims import LimsAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants.constants import CancerAnalysisType, CustomerId, Workflow
from cg.constants.observations import LoqusdbInstance, MipDNALoadParameters
from cg.constants.sample_sources import SourceType
from cg.constants.sequencing import SequencingMethod
from cg.exc import CaseNotFoundError, LoqusdbUploadCaseError
from cg.meta.observations.observations_api import ObservationsAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import ObservationsInputFiles
from cg.store.models import Case, Customer


@pytest.mark.parametrize(
    "workflow, analysis_api, sequencing_method, is_eligible, expected_message",
    [
        (
            Workflow.BALSAMIC,
            BalsamicAnalysisAPI,
            CancerAnalysisType.TUMOR_WGS,
            True,
            "Uploaded {number_of_loaded_variants} variants to Loqusdb",
        ),
        (
            Workflow.MIP_DNA,
            MipDNAAnalysisAPI,
            SequencingMethod.WGS,
            True,
            "Uploaded {number_of_loaded_variants} variants to Loqusdb",
        ),
        (
            Workflow.BALSAMIC,
            BalsamicAnalysisAPI,
            CancerAnalysisType.TUMOR_WGS,
            False,
            "Case {case_id} is not eligible for observations upload",
        ),
        (
            Workflow.MIP_DNA,
            MipDNAAnalysisAPI,
            SequencingMethod.WGS,
            False,
            "Case {case_id} is not eligible for observations upload",
        ),
    ],
)
def test_observations_upload(
    case_id: str,
    loqusdb_id: str,
    number_of_loaded_variants: int,
    workflow: Workflow,
    analysis_api: AnalysisAPI,
    sequencing_method: str,
    is_eligible: bool,
    expected_message: str,
    request: FixtureRequest,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test upload of observations."""
    caplog.set_level(logging.DEBUG)

    # GIVEN an observations API and a list of observation input files
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    observations_input_files: ObservationsInputFiles = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_input_files"
    )

    # GIVEN a mock scenario for a successful upload
    mocker.patch.object(analysis_api, "get_data_analysis_type", return_value=sequencing_method)
    mocker.patch.object(
        ObservationsAPI, "get_observations_input_files", return_value=observations_input_files
    )
    mocker.patch.object(ObservationsAPI, "is_duplicate", return_value=False)

    # GIVEN a case not eligible for Loqusdb uploads
    if not is_eligible:
        mocker.patch.object(LimsAPI, "get_source", return_value=SourceType.TISSUE_FFPE)

    # WHEN uploading the case observations to Loqusdb
    if is_eligible:
        observations_api.upload(case_id)
    else:
        with pytest.raises(LoqusdbUploadCaseError):
            observations_api.upload(case_id)

    # THEN the expected log message should be present
    assert (
        expected_message.format(
            case_id=case_id, number_of_loaded_variants=number_of_loaded_variants
        )
        in caplog.text
    )


@pytest.mark.parametrize(
    "workflow, loqusdb_instance",
    [(Workflow.BALSAMIC, LoqusdbInstance.TUMOR), (Workflow.MIP_DNA, LoqusdbInstance.WES)],
)
def test_get_loqusdb_api(
    cg_context: CGConfig,
    workflow: Workflow,
    loqusdb_instance: LoqusdbInstance,
    request: FixtureRequest,
):
    """Test Loqusdb API retrieval given a Loqusdb instance."""

    # GIVEN a WES Loqusdb instance and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )

    # GIVEN the expected Loqusdb config dictionary
    loqusdb_instance_key: str = loqusdb_instance.value.replace("-", "_")
    loqusdb_config: dict[str, Path] = cg_context.dict()[loqusdb_instance_key]

    # WHEN calling the Loqusdb API get method
    loqusdb_api: LoqusdbAPI = observations_api.get_loqusdb_api(loqusdb_instance)

    # THEN a WES Loqusdb API should be returned
    assert isinstance(loqusdb_api, LoqusdbAPI)
    assert loqusdb_api.binary_path == loqusdb_config["binary_path"]
    assert loqusdb_api.config_path == loqusdb_config["config_path"]


@pytest.mark.parametrize(
    "workflow, loqusdb_instance",
    [(Workflow.BALSAMIC, LoqusdbInstance.TUMOR), (Workflow.MIP_DNA, LoqusdbInstance.WES)],
)
def test_is_not_duplicate(
    case_id: str,
    workflow: Workflow,
    loqusdb_instance: LoqusdbInstance,
    request: FixtureRequest,
    mocker: MockFixture,
):
    """Test duplicate extraction for a case that is not in Loqusdb."""

    # GIVEN an observations API and a list of files to upload
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    observations_input_files: ObservationsInputFiles = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_input_files"
    )

    # GIVEN a Loqusdb instance with no case duplicates
    case: Case = observations_api.store.get_case_by_internal_id(case_id)
    loqusdb_api: LoqusdbAPI = observations_api.get_loqusdb_api(loqusdb_instance)
    mocker.patch.object(LoqusdbAPI, "get_case", return_value=None)
    mocker.patch.object(LoqusdbAPI, "get_duplicate", return_value=False)

    # WHEN checking that case has not been uploaded to Loqusdb
    is_duplicate: bool = observations_api.is_duplicate(
        case=case,
        loqusdb_api=loqusdb_api,
        profile_vcf_path=observations_input_files.snv_vcf_path,
        profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
    )

    # THEN there should be no duplicates in Loqusdb
    assert is_duplicate is False


@pytest.mark.parametrize(
    "workflow, loqusdb_instance",
    [(Workflow.BALSAMIC, LoqusdbInstance.TUMOR), (Workflow.MIP_DNA, LoqusdbInstance.WES)],
)
def test_is_duplicate(
    case_id: str,
    workflow: Workflow,
    loqusdb_instance: LoqusdbInstance,
    request: FixtureRequest,
    mocker: MockFixture,
):
    """Test duplicate extraction for a case that already exists in Loqusdb."""

    # GIVEN an observations API and a list of files to upload
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    observations_input_files: ObservationsInputFiles = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_input_files"
    )

    # GIVEN a Loqusdb instance with a duplicated case
    case: Case = observations_api.store.get_case_by_internal_id(case_id)
    loqusdb_api: LoqusdbAPI = observations_api.get_loqusdb_api(loqusdb_instance)
    mocker.patch.object(LoqusdbAPI, "get_case", return_value=None)
    mocker.patch.object(LoqusdbAPI, "get_duplicate", return_value={"case_id": case_id})

    # WHEN checking that case has not been uploaded to Loqusdb
    is_duplicate: bool = observations_api.is_duplicate(
        case=case,
        loqusdb_api=loqusdb_api,
        profile_vcf_path=observations_input_files.snv_vcf_path,
        profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
    )

    # THEN an upload of a duplicate case should be detected
    assert is_duplicate is True


@pytest.mark.parametrize(
    "workflow, loqusdb_instance",
    [(Workflow.BALSAMIC, LoqusdbInstance.TUMOR), (Workflow.MIP_DNA, LoqusdbInstance.WES)],
)
def test_is_duplicate_loqusdb_id(
    case_id: str,
    loqusdb_id: str,
    workflow: Workflow,
    loqusdb_instance: LoqusdbInstance,
    request: FixtureRequest,
    mocker: MockFixture,
):
    """Test duplicate extraction for a case that already exists in Loqusdb."""

    # GIVEN an observations API and a list of files to upload
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    observations_input_files: ObservationsInputFiles = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_input_files"
    )

    # GIVEN a Loqusdb instance with a duplicated case and whose samples already have a Loqusdb ID
    case: Case = observations_api.store.get_case_by_internal_id(case_id)
    loqusdb_api: LoqusdbAPI = observations_api.get_loqusdb_api(loqusdb_instance)
    case.links[0].sample.loqusdb_id = loqusdb_id
    mocker.patch.object(LoqusdbAPI, "get_case", return_value=None)
    mocker.patch.object(LoqusdbAPI, "get_duplicate", return_value=False)

    # WHEN checking that the sample observations have already been uploaded
    is_duplicate: bool = observations_api.is_duplicate(
        case=case,
        loqusdb_api=loqusdb_api,
        profile_vcf_path=observations_input_files.snv_vcf_path,
        profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
    )

    # THEN a duplicated upload should be identified
    assert is_duplicate is True


@pytest.mark.parametrize("workflow", [Workflow.BALSAMIC, Workflow.MIP_DNA])
def test_is_customer_eligible_for_observations_upload(
    workflow: Workflow,
    request: FixtureRequest,
):
    """Test if customer is eligible for observations upload."""

    # GIVEN a MIP-DNA customer and observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    customer: Customer = request.getfixturevalue(f"{workflow.replace('-', '_')}_customer")
    customer_id: str = customer.internal_id

    # WHEN verifying if the customer is eligible for Balsamic observations upload
    is_customer_eligible_for_observations_upload: bool = (
        observations_api.is_customer_eligible_for_observations_upload(customer_id)
    )

    # THEN the customer's data should be eligible for uploads
    assert is_customer_eligible_for_observations_upload


@pytest.mark.parametrize("workflow", [Workflow.BALSAMIC, Workflow.MIP_DNA])
def test_is_customer_not_eligible_for_observations_upload(
    workflow: Workflow, request: FixtureRequest, caplog: LogCaptureFixture
):
    """Test if customer is not eligible for observations upload."""

    # GIVEN a CG internal customer ID and observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    customer_id: str = CustomerId.CG_INTERNAL_CUSTOMER

    # WHEN verifying if the customer is eligible for Balsamic observations upload
    is_customer_eligible_for_observations_upload: bool = (
        observations_api.is_customer_eligible_for_observations_upload(customer_id)
    )

    # THEN the customer's data should not be eligible for uploads
    assert not is_customer_eligible_for_observations_upload
    assert f"Customer {customer_id} is not whitelisted for Loqusdb uploads" in caplog.text


@pytest.mark.parametrize(
    "workflow, analysis_api, sequencing_method",
    [
        (Workflow.BALSAMIC, BalsamicAnalysisAPI, CancerAnalysisType.TUMOR_WGS),
        (Workflow.MIP_DNA, MipDNAAnalysisAPI, SequencingMethod.WGS),
    ],
)
def test_is_sequencing_method_eligible_for_observations_upload(
    case_id: str,
    workflow: Workflow,
    analysis_api: AnalysisAPI,
    sequencing_method: str,
    request: FixtureRequest,
    mocker: MockFixture,
):
    """Test if the sequencing method is eligible for observations uploads."""

    # GIVEN a case ID and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )

    # GIVEN a supported data analysis type
    mocker.patch.object(analysis_api, "get_data_analysis_type", return_value=sequencing_method)

    # WHEN verifying that the sequencing method is eligible for observations uploads
    is_sequencing_method_eligible_for_observations_upload: bool = (
        observations_api.is_sequencing_method_eligible_for_observations_upload(case_id)
    )

    # THEN the sequencing method should be eligible for observations uploads
    assert is_sequencing_method_eligible_for_observations_upload


@pytest.mark.parametrize(
    "workflow, analysis_api, sequencing_method",
    [
        (Workflow.BALSAMIC, BalsamicAnalysisAPI, CancerAnalysisType.TUMOR_PANEL),
        (Workflow.MIP_DNA, MipDNAAnalysisAPI, SequencingMethod.WTS),
    ],
)
def test_is_sequencing_method_not_eligible_for_observations_upload(
    case_id: str,
    workflow: Workflow,
    analysis_api: AnalysisAPI,
    sequencing_method: str,
    request: FixtureRequest,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test if the sequencing method is eligible for observations uploads."""

    # GIVEN a case ID and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )
    # GIVEN a non-supported data analysis type
    mocker.patch.object(analysis_api, "get_data_analysis_type", return_value=sequencing_method)

    # WHEN verifying that the sequencing method is eligible for observations uploads
    is_sequencing_method_eligible_for_observations_upload: bool = (
        observations_api.is_sequencing_method_eligible_for_observations_upload(case_id)
    )

    # THEN the sequencing method should not be eligible for observations uploads
    assert not is_sequencing_method_eligible_for_observations_upload
    assert (
        f"Sequencing method {sequencing_method} is not supported by Loqusdb uploads" in caplog.text
    )


@pytest.mark.parametrize("workflow", [Workflow.BALSAMIC, Workflow.MIP_DNA])
def test_is_sample_source_eligible_for_observations_upload(
    case_id: str, workflow: Workflow, request: FixtureRequest, mocker: MockFixture
):
    """Test if the sample source is eligible for observations uploads."""

    # GIVEN a case ID and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )

    # GIVEN a supported sample source

    # WHEN verifying that the sample source is eligible for observations uploads
    is_sample_source_eligible_for_observations_upload: bool = (
        observations_api.is_sample_source_eligible_for_observations_upload(case_id)
    )

    # THEN the source type should be eligible for observations uploads
    assert is_sample_source_eligible_for_observations_upload


@pytest.mark.parametrize("workflow", [Workflow.BALSAMIC, Workflow.MIP_DNA])
def test_is_sample_source_not_eligible_for_observations_upload(
    case_id: str,
    workflow: Workflow,
    request: FixtureRequest,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test if the sample source is not eligible for observations uploads."""

    # GIVEN a case ID and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )

    # GIVEN a not supported sample source
    source_type = SourceType.TISSUE_FFPE
    mocker.patch.object(LimsAPI, "get_source", return_value=source_type)

    # WHEN verifying that the sample source is eligible for observations uploads
    is_sample_source_eligible_for_observations_upload: bool = (
        observations_api.is_sample_source_eligible_for_observations_upload(case_id)
    )

    # THEN the source type should not be eligible for observations uploads
    assert not is_sample_source_eligible_for_observations_upload
    assert f"Source type {source_type} is not supported for Loqusdb uploads" in caplog.text


@pytest.mark.parametrize(
    "workflow, analysis_api, sequencing_method",
    [
        (Workflow.BALSAMIC, BalsamicAnalysisAPI, CancerAnalysisType.TUMOR_WGS),
        (Workflow.MIP_DNA, MipDNAAnalysisAPI, SequencingMethod.WGS),
    ],
)
def test_delete_case(
    case_id: str,
    loqusdb_id: str,
    workflow: Workflow,
    analysis_api: AnalysisAPI,
    sequencing_method: str,
    request: FixtureRequest,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test delete case from Loqusdb."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a case ID and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )

    # GIVEN a case uploaded to Loqusdb
    mocker.patch.object(analysis_api, "get_data_analysis_type", return_value=sequencing_method)
    mocker.patch.object(LoqusdbAPI, "delete_case", return_value=None)

    # WHEN deleting a case
    observations_api.delete_case(case_id)

    # THEN the case should be deleted from Loqusdb
    assert f"Removed observations for case {case_id}" in caplog.text


@pytest.mark.parametrize(
    "workflow, analysis_api, sequencing_method",
    [
        (Workflow.BALSAMIC, BalsamicAnalysisAPI, CancerAnalysisType.TUMOR_WGS),
        (Workflow.MIP_DNA, MipDNAAnalysisAPI, SequencingMethod.WGS),
    ],
)
def test_delete_case_not_found(
    case_id: str,
    loqusdb_id: str,
    workflow: Workflow,
    analysis_api: AnalysisAPI,
    sequencing_method: str,
    request: FixtureRequest,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test delete case from Loqusdb that has not been uploaded."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a case ID and an observations API
    observations_api: ObservationsAPI = request.getfixturevalue(
        f"{workflow.replace('-', '_')}_observations_api"
    )

    # GIVEN a case that has not been uploaded to Loqusdb
    mocker.patch.object(LoqusdbAPI, "get_case", return_value=None)
    mocker.patch.object(analysis_api, "get_data_analysis_type", return_value=sequencing_method)

    # WHEN deleting a case
    with pytest.raises(CaseNotFoundError):
        observations_api.delete_case(case_id)

    # THEN a CaseNotFoundError should be raised
    assert f"Case {case_id} could not be found in Loqusdb. Skipping case deletion." in caplog.text



"""API for uploading cancer observations."""

import logging
from pathlib import Path

from housekeeper.store.models import File, Version

from cg.apps.loqus import LoqusdbAPI
from cg.constants.constants import CancerAnalysisType, CustomerId
from cg.constants.observations import (
    LOQUSDB_CANCER_CUSTOMERS,
    LOQUSDB_CANCER_SEQUENCING_METHODS,
    LOQUSDB_ID,
    BalsamicLoadParameters,
    BalsamicObservationsAnalysisTag,
    LoqusdbInstance,
)
from cg.exc import CaseNotFoundError, LoqusdbDuplicateRecordError
from cg.meta.observations.observations_api import ObservationsAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import BalsamicObservationsInputFiles
from cg.store.models import Case
from cg.utils.dict import get_full_path_dictionary

LOG = logging.getLogger(__name__)


class BalsamicObservationsAPI(ObservationsAPI):
    """API to manage Balsamic observations."""

    def __init__(self, config: CGConfig):
        self.analysis_api = BalsamicAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)
        self.loqusdb_somatic_api: LoqusdbAPI = self.get_loqusdb_api(LoqusdbInstance.SOMATIC)
        self.loqusdb_tumor_api: LoqusdbAPI = self.get_loqusdb_api(LoqusdbInstance.TUMOR)

    @property
    def loqusdb_customers(self) -> list[CustomerId]:
        """Customers that are eligible for cancer Loqusdb uploads."""
        return LOQUSDB_CANCER_CUSTOMERS

    @property
    def loqusdb_sequencing_methods(self) -> list[str]:
        """Return sequencing methods that are eligible for cancer Loqusdb uploads."""
        return LOQUSDB_CANCER_SEQUENCING_METHODS

    def is_analysis_type_eligible_for_observations_upload(self, case_id) -> bool:
        """Return whether the cancer analysis type is eligible for cancer Loqusdb uploads."""
        if self.analysis_api.is_analysis_normal_only(case_id):
            LOG.error(f"Normal only analysis {case_id} is not supported for Loqusdb uploads")
            return False
        return True

    def is_case_eligible_for_observations_upload(self, case: Case) -> bool:
        """Return whether a cancer case is eligible for observations upload."""
        return all(
            [
                self.is_customer_eligible_for_observations_upload(case.customer.internal_id),
                self.is_sequencing_method_eligible_for_observations_upload(case.internal_id),
                self.is_analysis_type_eligible_for_observations_upload(case.internal_id),
                self.is_sample_source_eligible_for_observations_upload(case.internal_id),
            ]
        )

    def load_observations(self, case: Case) -> None:
        """
        Load observation counts to Loqusdb for a Balsamic case.

        Raises:
            LoqusdbDuplicateRecordError: If case has already been uploaded.
        """
        loqusdb_upload_apis: list[LoqusdbAPI] = [self.loqusdb_somatic_api, self.loqusdb_tumor_api]
        for loqusdb_api in loqusdb_upload_apis:
            if self.is_duplicate(case=case, loqusdb_api=loqusdb_api):
                LOG.error(f"Case {case.internal_id} has already been uploaded to Loqusdb")
                raise LoqusdbDuplicateRecordError

        input_files: BalsamicObservationsInputFiles = self.get_observations_input_files(case)
        for loqusdb_api in loqusdb_upload_apis:
            self.load_cancer_observations(
                case=case, input_files=input_files, loqusdb_api=loqusdb_api
            )

        # Update Statusdb with a germline Loqusdb ID
        loqusdb_id: str = str(self.loqusdb_tumor_api.get_case(case_id=case.internal_id)[LOQUSDB_ID])
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=loqusdb_id)

    def load_cancer_observations(
        self, case: Case, input_files: BalsamicObservationsInputFiles, loqusdb_api: LoqusdbAPI
    ) -> None:
        """Load cancer observations to a specific Loqusdb API."""
        is_somatic_db: bool = LoqusdbInstance.SOMATIC in str(loqusdb_api.config_path)
        is_paired_analysis: bool = (
            CancerAnalysisType.TUMOR_NORMAL
            in self.analysis_api.get_data_analysis_type(case.internal_id)
        )
        if is_somatic_db:
            if not is_paired_analysis:
                return
            LOG.info("Uploading somatic observations to Loqusdb")
            snv_vcf_path: Path = input_files.snv_vcf_path
            sv_vcf_path: Path = input_files.sv_vcf_path
        else:
            LOG.info("Uploading germline observations to Loqusdb")
            snv_vcf_path: Path = input_files.snv_germline_vcf_path
            sv_vcf_path: Path = input_files.sv_germline_vcf_path if is_paired_analysis else None

        load_output: dict = loqusdb_api.load(
            case_id=case.internal_id,
            snv_vcf_path=snv_vcf_path,
            sv_vcf_path=sv_vcf_path,
            qual_gq=True,
            gq_threshold=(
                BalsamicLoadParameters.QUAL_THRESHOLD.value
                if is_somatic_db
                else BalsamicLoadParameters.QUAL_GERMLINE_THRESHOLD.value
            ),
        )
        LOG.info(f"Uploaded {load_output['variants']} variants to {repr(loqusdb_api)}")

    def get_observations_files_from_hk(
        self, hk_version: Version, case_id: str = None
    ) -> BalsamicObservationsInputFiles:
        """Return observations files given a Housekeeper version for cancer."""
        input_files: dict[str, File] = {
            "snv_germline_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SNV_GERMLINE_VCF]
            ).first(),
            "snv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SNV_VCF]
            ).first(),
            "sv_germline_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SV_GERMLINE_VCF]
            ).first(),
            "sv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SV_VCF]
            ).first(),
        }
        return BalsamicObservationsInputFiles(**get_full_path_dictionary(input_files))

    def delete_case(self, case_id: str) -> None:
        """Delete cancer case observations from Loqusdb."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        loqusdb_apis: list[LoqusdbAPI] = [self.loqusdb_somatic_api, self.loqusdb_tumor_api]
        for loqusdb_api in loqusdb_apis:
            if not loqusdb_api.get_case(case_id):
                LOG.error(f"Case {case_id} could not be found in Loqusdb. Skipping case deletion.")
                raise CaseNotFoundError
        for loqusdb_api in loqusdb_apis:
            loqusdb_api.delete_case(case_id)
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=None)
        LOG.info(f"Removed observations for case {case_id} from Loqusdb")



"""Tests for the sequencing metrics dtos."""

from cg.server.utils import parse_metrics_into_request
from cg.store.models import IlluminaSampleSequencingMetrics
from cg.store.store import Store


def test_sequencing_metrics_request(
    store_with_illumina_sequencing_data: Store, novaseq_x_flow_cell_id: str
):
    # GIVEN some sequencing metrics
    sequencing_metrics: list[IlluminaSampleSequencingMetrics] = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        ).sample_metrics
    )
    assert sequencing_metrics

    # WHEN parsing the sequencing metrics into a request
    parsed_metrics = parse_metrics_into_request(sequencing_metrics)

    # THEN the parsed metrics should be equal to the original metrics
    for pos in range(1, len(sequencing_metrics)):
        assert (
            sequencing_metrics[pos].instrument_run.device.internal_id
            == parsed_metrics[pos].flow_cell_name
        )
        assert sequencing_metrics[pos].flow_cell_lane == parsed_metrics[pos].flow_cell_lane_number
        assert sequencing_metrics[pos].sample.internal_id == parsed_metrics[pos].sample_internal_id
        assert (
            sequencing_metrics[pos].total_reads_in_lane
            == parsed_metrics[pos].sample_total_reads_in_lane
        )



"""Module for PacBio fixtures returning strings."""

import pytest

from cg.constants.pacbio import PacBioDirsAndFiles


@pytest.fixture
def pac_bio_test_run_name() -> str:
    """Return the name of a PacBio SMRT cell."""
    return "r84202_20240522_133539"


@pytest.fixture
def pac_bio_1_a01_cell_full_name() -> str:
    """Return the full name of a PacBio SMRT cell."""
    return "m84202_240522_135641_s1"


@pytest.fixture
def ccs_report_1_a01_name(pac_bio_1_a01_cell_full_name: str) -> str:
    """Return the name of a ccs report file."""
    return f"{pac_bio_1_a01_cell_full_name}.{PacBioDirsAndFiles.CCS_REPORT_SUFFIX}"



"""CLI function to compress FASTQ files into SPRING archives."""

import logging
from typing import Iterable

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.compress.helpers import (
    compress_sample_fastqs_in_cases,
    correct_spring_paths,
    get_cases_to_process,
    update_compress_api,
)
from cg.constants.cli_options import DRY_RUN
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.command("fastq")
@click.option("-c", "--case-id", type=str)
@click.option(
    "-b",
    "--days-back",
    default=60,
    show_default=True,
    help="Threshold for how long ago was the case created",
)
@click.option("--hours", type=int, help="Hours to allocate for slurm job")
@click.option("-m", "--mem", type=int, help="Memory for slurm job")
@click.option("-t", "--ntasks", type=int, help="Number of tasks for slurm job")
@click.option("-n", "--number-of-conversions", default=5, type=int, show_default=True)
@DRY_RUN
@click.pass_obj
def fastq_cmd(
    context: CGConfig,
    case_id: str | None,
    days_back: int,
    hours: int | None,
    dry_run: bool,
    mem: int | None,
    ntasks: int | None,
    number_of_conversions: int,
):
    """Compress old FASTQ files into SPRING."""
    LOG.info("Running compress FASTQ")
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    store: Store = context.status_db
    cases: list[Case] = get_cases_to_process(case_id=case_id, days_back=days_back, store=store)
    if not cases:
        LOG.info("No cases to compress")
        return None
    compress_sample_fastqs_in_cases(
        compress_api=compress_api,
        cases=cases,
        dry_run=dry_run,
        number_of_conversions=number_of_conversions,
        hours=hours,
        mem=mem,
        ntasks=ntasks,
    )


@click.command("fastq")
@click.option("-c", "--case-id")
@click.option(
    "-b",
    "--days-back",
    default=60,
    show_default=True,
    help="Threshold for how long ago was the case created",
)
@DRY_RUN
@click.pass_obj
def clean_fastq(context: CGConfig, case_id: str | None, days_back: int, dry_run: bool):
    """Remove compressed FASTQ files, and update links in Housekeeper to SPRING files."""
    LOG.info("Running compress clean FASTQ")
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    store: Store = context.status_db
    update_compress_api(compress_api, dry_run=dry_run)

    cases: list[Case] = get_cases_to_process(case_id=case_id, days_back=days_back, store=store)
    if not cases:
        return

    cleaned_inds = 0
    for case in cases:
        sample_ids: Iterable[str] = store.get_sample_ids_by_case_id(case_id=case.internal_id)
        for sample_id in sample_ids:
            archive_location: str = store.get_sample_by_internal_id(sample_id).archive_location
            was_cleaned: bool = compress_api.clean_fastq(
                sample_id=sample_id, archive_location=archive_location
            )
            if not was_cleaned:
                LOG.info(f"Skipping individual {sample_id}")
                continue
            cleaned_inds += 1

    LOG.info(f"Cleaned fastqs in {cleaned_inds} individuals")


@click.command("fix-spring")
@click.option("-b", "--bundle-name")
@DRY_RUN
@click.pass_obj
def fix_spring(context: CGConfig, bundle_name: str | None, dry_run: bool):
    """Check if bundle(s) have non-existing SPRING files and correct these."""
    LOG.info("Running fix spring")
    compress_api = context.meta_apis["compress_api"]
    update_compress_api(compress_api, dry_run=dry_run)
    hk_api: HousekeeperAPI = compress_api.hk_api
    correct_spring_paths(hk_api=hk_api, bundle_name=bundle_name, dry_run=dry_run)


@click.command("sample")
@click.argument("sample-id", type=str)
@DRY_RUN
@click.pass_obj
def decompress_sample(context: CGConfig, sample_id: str, dry_run: bool):
    """Decompress SPRING file for sample, and include links to FASTQ files in Housekeeper."""

    compress_api: CompressAPI = context.meta_apis["compress_api"]
    update_compress_api(compress_api=compress_api, dry_run=dry_run)

    was_decompressed: bool = compress_api.decompress_spring(sample_id=sample_id)
    if not was_decompressed:
        LOG.info(f"Skipping sample {sample_id}")
        return 0
    LOG.info(f"Decompressed sample {sample_id}")
    return 1


@click.command("case")
@click.argument("case-id", type=str)
@DRY_RUN
@click.pass_context
def decompress_case(context: click.Context, case_id, dry_run):
    """Decompress SPRING file for case, and include links to FASTQ files in Housekeeper."""

    store: Store = context.obj.status_db
    try:
        samples: Iterable[str] = store.get_sample_ids_by_case_id(case_id=case_id)
        decompressed_individuals = 0
        for sample_id in samples:
            decompressed_count: int = context.invoke(
                decompress_sample, sample_id=sample_id, dry_run=dry_run
            )
            decompressed_individuals += decompressed_count
    except CaseNotFoundError:
        return
    LOG.info(f"Decompressed spring archives in {decompressed_individuals} samples")


@click.command("illumina-run")
@click.argument("flow-cell-id", type=str)
@DRY_RUN
@click.pass_obj
def decompress_illumina_run(context: click.Context, flow_cell_id: str, dry_run: bool):
    """Decompress SPRING files for flow cell, and include links to FASTQ files in Housekeeper."""

    store: Store = context.obj.status_db
    samples: Iterable[Sample] = store.get_samples_by_illumina_flow_cell(flow_cell_id)
    decompressed_individuals = 0
    for sample in samples:
        decompressed_count = context.invoke(
            decompress_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        decompressed_individuals += decompressed_count
    LOG.info(f"Decompressed spring archives in {decompressed_individuals} samples")


@click.command("ticket")
@click.argument("ticket", type=str)
@DRY_RUN
@click.pass_context
def decompress_ticket(context: click.Context, ticket: str, dry_run: bool):
    """Decompress SPRING file for ticket, and include links to FASTQ files in Housekeeper."""
    store: Store = context.obj.status_db
    samples: Iterable[Sample] = store.get_samples_from_ticket(ticket=ticket)
    decompressed_individuals = 0
    for sample in samples:
        decompressed_count = context.invoke(
            decompress_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        decompressed_individuals += decompressed_count
    LOG.info(f"Decompressed spring archives in {decompressed_individuals} samples")



"""Tests for the ValidatePacBioFileTransferService class."""

from pathlib import Path

from cg.constants.file_transfer_service import PACBIO_MANIFEST_FILE_PATTERN, TRANSFER_VALIDATED_FILE
from cg.models.cg_config import CGConfig
from cg.services.validate_file_transfer_service.validate_pacbio_file_transfer_service import (
    ValidatePacbioFileTransferService,
)


def test_create_systemd_trigger_file(
    pacbio_runs_dir: Path,
    pacbio_run_id: str,
    validate_pacbio_service: ValidatePacbioFileTransferService,
    smrt_cell_id: str,
):
    """Test the create systemd trigger file method."""
    # GIVEN a manifest file

    manifest_file: Path = validate_pacbio_service.get_manifest_file_paths(
        source_dir=pacbio_runs_dir, pattern=PACBIO_MANIFEST_FILE_PATTERN
    )[0]

    # WHEN creating the systemd trigger file
    validate_pacbio_service.create_systemd_trigger_file(manifest_file_path=manifest_file)

    # THEN assert that the systemd trigger file was created
    assert Path(pacbio_runs_dir, pacbio_run_id + "-" + smrt_cell_id).exists()


def test_create_validated_transfer_file(
    pacbio_runs_dir: Path,
    pacbio_run_id: str,
    validate_pacbio_service: ValidatePacbioFileTransferService,
):
    """Test the create validated transfer file method."""
    # GIVEN a manifest file

    manifest_file: Path = validate_pacbio_service.get_manifest_file_paths(
        source_dir=pacbio_runs_dir, pattern=PACBIO_MANIFEST_FILE_PATTERN
    )[0]

    # WHEN creating the validated transfer file
    validate_pacbio_service.create_validated_transfer_file(manifest_file_path=manifest_file)

    # THEN assert that the validated transfer file was created
    assert Path(manifest_file.parent, TRANSFER_VALIDATED_FILE).exists()


def test_get_run_id(
    pacbio_runs_dir: Path,
    pacbio_run_id: str,
    validate_pacbio_service: ValidatePacbioFileTransferService,
):
    """Test the get run ID method."""
    # GIVEN a manifest file

    manifest_file: Path = validate_pacbio_service.get_manifest_file_paths(
        source_dir=pacbio_runs_dir, pattern=PACBIO_MANIFEST_FILE_PATTERN
    )[0]

    # WHEN getting the run ID
    extracted_run_id: str = validate_pacbio_service.get_run_id(manifest_file_path=manifest_file)

    # THEN it is the expected run id
    assert extracted_run_id == pacbio_run_id


def test_get_smrt_cell_id(
    pacbio_runs_dir: Path,
    smrt_cell_id: str,
    validate_pacbio_service: ValidatePacbioFileTransferService,
):
    """Test the get smrt cell ID method."""
    # GIVEN a manifest file

    manifest_file: Path = validate_pacbio_service.get_manifest_file_paths(
        source_dir=pacbio_runs_dir, pattern=PACBIO_MANIFEST_FILE_PATTERN
    )[0]

    # WHEN getting the smrt cell ID
    extracted_smrt_cell_id: str = validate_pacbio_service.get_smrt_cell_id(
        manifest_file_path=manifest_file
    )

    # THEN it is the expected smrt cell id
    assert extracted_smrt_cell_id == smrt_cell_id



"""Module for PacBio fixtures returning service objects."""

from pathlib import Path

import pytest

from cg.services.pacbio.metrics.metrics_parser import PacBioMetricsParser


@pytest.fixture
def pac_bio_metrics_parser(pac_bio_smrt_cell_dir_1_a01: Path) -> PacBioMetricsParser:
    """Return a PacBio metrics parser."""
    return PacBioMetricsParser(pac_bio_smrt_cell_dir_1_a01)



"""Chanjo2 API model fixtures."""

import pytest

from cg.clients.chanjo2.models import CoveragePostRequest, CoveragePostResponse, CoverageSample
from cg.constants.constants import GenomeVersion


@pytest.fixture
def coverage_post_request(sample_id: str) -> CoveragePostRequest:
    return CoveragePostRequest(
        build=GenomeVersion.GRCh37,
        coverage_threshold=10,
        hgnc_gene_ids=[2861, 3791, 6481, 7436, 30521],
        interval_type="genes",
        samples=[
            CoverageSample(
                coverage_file_path=f"/path/to/coverage/{sample_id}.d4",
                name=sample_id,
            )
        ],
    )


@pytest.fixture
def coverage_post_response(coverage_post_response_json: dict) -> CoveragePostResponse:
    return CoveragePostResponse.model_validate(coverage_post_response_json)



"""CLI options for Nextflow and NF-Tower."""

import logging

import click
from pydantic import ValidationError

from cg.cli.workflow.commands import ARGUMENT_CASE_ID
from cg.cli.workflow.utils import validate_force_store_option
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.constants.cli_options import DRY_RUN, FORCE, COMMENT
from cg.constants.constants import MetaApis
from cg.exc import AnalysisNotReadyError, CgError, HousekeeperStoreError
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)

OPTION_WORKDIR = click.option(
    "--work-dir",
    type=click.Path(),
    help="Directory where intermediate result files are stored",
)
OPTION_RESUME = click.option(
    "--resume",
    is_flag=True,
    default=False,
    show_default=True,
    help="Execute the script using the cached results, useful to continue \
        executions that was stopped by an error",
)
OPTION_PROFILE = click.option(
    "--profile",
    type=str,
    show_default=True,
    help="Choose a configuration profile",
)

OPTION_LOG = click.option(
    "--log",
    type=click.Path(),
    help="Set nextflow log file path",
)

OPTION_CONFIG = click.option(
    "--config",
    type=click.Path(),
    help="Nextflow config file path",
)

OPTION_PARAMS_FILE = click.option(
    "--params-file",
    type=click.Path(),
    help="Nextflow workflow-specific parameter file path",
)

OPTION_USE_NEXTFLOW = click.option(
    "--use-nextflow",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="Execute workflow using nextflow",
)

OPTION_REVISION = click.option(
    "--revision",
    type=str,
    help="Revision of workflow to run (either a git branch, tag or commit SHA number)",
)
OPTION_COMPUTE_ENV = click.option(
    "--compute-env",
    type=str,
    help="Compute environment name. If not specified the primary compute environment will be used.",
)
OPTION_TOWER_RUN_ID = click.option(
    "--nf-tower-id",
    type=str,
    is_flag=False,
    default=None,
    help="NF-Tower ID of run to relaunch. If not provided the latest NF-Tower ID for a case will be used.",
)
OPTION_FROM_START = click.option(
    "--from-start",
    is_flag=True,
    default=False,
    show_default=True,
    help="Start workflow from start without resuming execution",
)


@click.command("config-case")
@ARGUMENT_CASE_ID
@DRY_RUN
@click.pass_obj
def config_case(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Create config files required by a workflow for a case."""
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.config_case(case_id=case_id, dry_run=dry_run)
    except (CgError, ValidationError) as error:
        LOG.error(f"Could not create config files for {case_id}: {error}")
        raise click.Abort() from error


@click.command("run")
@ARGUMENT_CASE_ID
@OPTION_LOG
@OPTION_WORKDIR
@OPTION_FROM_START
@OPTION_PROFILE
@OPTION_CONFIG
@OPTION_PARAMS_FILE
@OPTION_REVISION
@OPTION_COMPUTE_ENV
@OPTION_USE_NEXTFLOW
@OPTION_TOWER_RUN_ID
@DRY_RUN
@click.pass_obj
def run(
    context: CGConfig,
    case_id: str,
    log: str,
    work_dir: str,
    from_start: bool,
    profile: str,
    config: str,
    params_file: str,
    revision: str,
    compute_env: str,
    use_nextflow: bool,
    nf_tower_id: str | None,
    dry_run: bool,
) -> None:
    """Run analysis for a case."""
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.run_nextflow_analysis(
            case_id=case_id,
            dry_run=dry_run,
            log=log,
            work_dir=work_dir,
            from_start=from_start,
            profile=profile,
            config=config,
            params_file=params_file,
            revision=revision,
            compute_env=compute_env,
            use_nextflow=use_nextflow,
            nf_tower_id=nf_tower_id,
        )
    except Exception as error:
        LOG.error(f"Unspecified error occurred: {error}")
        raise click.Abort() from error


@click.command("start")
@ARGUMENT_CASE_ID
@OPTION_LOG
@OPTION_WORKDIR
@OPTION_PROFILE
@OPTION_CONFIG
@OPTION_PARAMS_FILE
@OPTION_REVISION
@OPTION_COMPUTE_ENV
@OPTION_USE_NEXTFLOW
@DRY_RUN
@click.pass_obj
def start(
    context: CGConfig,
    case_id: str,
    log: str,
    work_dir: str,
    profile: str,
    config: str,
    params_file: str,
    revision: str,
    compute_env: str,
    use_nextflow: bool,
    dry_run: bool,
) -> None:
    """Start workflow for a case."""
    LOG.info(f"Starting analysis for {case_id}")
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.prepare_fastq_files(case_id=case_id, dry_run=dry_run)
        analysis_api.config_case(case_id=case_id, dry_run=dry_run)
        analysis_api.run_nextflow_analysis(
            case_id=case_id,
            dry_run=dry_run,
            log=log,
            work_dir=work_dir,
            from_start=True,
            profile=profile,
            config=config,
            params_file=params_file,
            revision=revision,
            compute_env=compute_env,
            use_nextflow=use_nextflow,
        )
    except Exception as error:
        LOG.error(f"Unexpected error occurred: {error}")
        raise click.Abort from error


@click.command("start-available")
@DRY_RUN
@click.pass_context
def start_available(context: click.Context, dry_run: bool = False) -> None:
    """Start workflow for all cases ready for analysis."""
    analysis_api: NfAnalysisAPI = context.obj.meta_apis[MetaApis.ANALYSIS_API]
    exit_code: int = EXIT_SUCCESS
    for case in analysis_api.get_cases_ready_for_analysis():
        try:
            context.invoke(start, case_id=case.internal_id, dry_run=dry_run)
        except AnalysisNotReadyError as error:
            LOG.error(error)
        except Exception as error:
            LOG.error(error)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort


@click.command("metrics-deliver")
@ARGUMENT_CASE_ID
@DRY_RUN
@click.pass_obj
def metrics_deliver(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """
    Create and validate a metrics deliverables file for given case id.
    If QC metrics are met it sets the status in Trailblazer to complete.
    If failed, it sets it as failed and adds a comment with information of the failed metrics.
    """
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.metrics_deliver(case_id=case_id, dry_run=dry_run)
    except CgError as error:
        raise click.Abort() from error


@click.command("report-deliver")
@ARGUMENT_CASE_ID
@DRY_RUN
@FORCE
@click.pass_obj
def report_deliver(context: CGConfig, case_id: str, dry_run: bool, force: bool) -> None:
    """
    Create a Housekeeper deliverables file for a given case ID.

    Raises:
        click.Abort: If an error occurs during the deliverables report generation or validation.
    """
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.report_deliver(case_id=case_id, dry_run=dry_run, force=force)
    except CgError as error:
        LOG.error(f"Could not create report file: {error}")
        raise click.Abort()


@click.command("store-housekeeper")
@ARGUMENT_CASE_ID
@DRY_RUN
@FORCE
@click.pass_obj
def store_housekeeper(context: CGConfig, case_id: str, dry_run: bool, force: bool) -> None:
    """
    Store a finished NF-analysis in Housekeeper.

    Raises:
        click.Abort: If an error occurs while storing a case bundle in Housekeeper.
    """
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.store_analysis_housekeeper(case_id=case_id, dry_run=dry_run, force=force)
    except HousekeeperStoreError as error:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {error}!")
        raise click.Abort()


@click.command("store")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_context
def store(
    context: click.Context, case_id: str, comment: str | None, dry_run: bool, force: bool
) -> None:
    """
    Store deliverable files in Housekeeper after meeting QC metrics criteria.

    Raises:
        click.Abort: If an error occurs during the deliverables file generation, metrics
        validation, or storage processes.
    """
    validate_force_store_option(force=force, comment=comment)
    analysis_api: NfAnalysisAPI = context.obj.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.store(case_id=case_id, comment=comment, dry_run=dry_run, force=force)
    except Exception as error:
        LOG.error(repr(error))
        raise click.Abort()


@click.command("store-available")
@DRY_RUN
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """
    Store finished analyses for cases marked as running in StatusDB and completed or QC in Trailblazer.

    Raises:
        click.Abort: If any error occurs during the storage process.
    """

    analysis_api: NfAnalysisAPI = context.obj.meta_apis[MetaApis.ANALYSIS_API]

    for case in analysis_api.get_cases_to_store():
        LOG.info(f"Storing deliverables for {case.internal_id}")
        try:
            analysis_api.store(case_id=case.internal_id, dry_run=dry_run)
        except Exception as error:
            LOG.error(f"Error storing {case.internal_id}: {repr(error)}")
            raise click.Abort



import logging
from functools import wraps
from http import HTTPStatus

import cachecontrol
import requests
from flask import abort, current_app, g, jsonify, make_response, request
from google.auth import exceptions
from google.auth.transport import requests as google_requests
from google.oauth2 import id_token


from cg.server.ext import db
from cg.store.models import User

LOG = logging.getLogger(__name__)

session = requests.session()
cached_session = cachecontrol.CacheControl(session)


def verify_google_token(token):
    request = google_requests.Request(session=cached_session)
    return id_token.verify_oauth2_token(id_token=token, request=request)


def is_public(route_function):
    @wraps(route_function)
    def public_endpoint(*args, **kwargs):
        return route_function(*args, **kwargs)

    public_endpoint.is_public = True
    return public_endpoint


def before_request():
    """Authorize API routes with JSON Web Tokens."""
    if not request.is_secure:
        return abort(
            make_response(jsonify(message="Only https requests accepted"), HTTPStatus.FORBIDDEN)
        )

    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), HTTPStatus.NO_CONTENT)

    endpoint_func = current_app.view_functions[request.endpoint]
    if getattr(endpoint_func, "is_public", None):
        return

    auth_header = request.headers.get("Authorization")
    if not auth_header:
        return abort(
            make_response(jsonify(message="no JWT token found on request"), HTTPStatus.UNAUTHORIZED)
        )

    jwt_token = auth_header.split("Bearer ")[-1]
    try:
        user_data = verify_google_token(jwt_token)
    except (exceptions.OAuthError, ValueError) as e:
        LOG.error(f"Error {e} occurred while decoding JWT token: {jwt_token}")
        return abort(
            make_response(jsonify(message="outdated login certificate"), HTTPStatus.UNAUTHORIZED)
        )

    user: User = db.get_user_by_email(user_data["email"])
    if user is None or not user.order_portal_login:
        message = f"{user_data['email']} doesn't have access"
        LOG.error(message)
        return abort(make_response(jsonify(message=message), HTTPStatus.FORBIDDEN))

    g.current_user = user
